//-------------------------------------------------------------
// particle_system.h:  particle_system class - a group of particles
//-------------------------------------------------------------
#ifndef BEAM_SYSTEM_H
#define BEAM_SYSTEM_H

//-------------------------------------------------------------
// includes / defines
//-------------------------------------------------------------

#include <d3dx9.h>				// DirectX 9
#include <vector>				// vector class
#include "BeamParticle.h"		// particle class
#include "Random.h"				// random number functions
using namespace std;

//-------------------------------------------------------------
//  beam_system - a particle system of beams
//-------------------------------------------------------------
class beam_system
{
	public:

	int num_beams;							// # beams in system
	D3DXVECTOR3 system_position;					// position of anchor
	vector<D3DXVECTOR3> position;			// position of beams
	vector<D3DXVECTOR3> base_color;			// origin color of beams
	vector<D3DXVECTOR3> end_color;			// end color of beams
	vector<beam_particle> particle_vector;	// list of particles
	vector<float> time_to_live;				// ttl each beam
	vector<float> life;						// timer for each beam
	vector<float> size;						// width of each beam

	beam_system();							// default contructor
	~beam_system();							// destructor

	void update(float time_elapsed, int index);			
	void draw();
	void reset(int index);
};

//-------------------------------------------------------------
//  beam_system::particle_system() - constructor
//-------------------------------------------------------------
beam_system::beam_system()
{
	int i;

	// set position
	system_position.x=0.0f;
	system_position.y=0.0f;
	system_position.z=0.0f;

	// add several beams
	num_beams = PARTICLE_DENSITY;

	for(i=0; i<num_beams; i++)
	{
		// position
		D3DXVECTOR3 pos(0.0f, 0.0f, 0.0f);	
		position.push_back(pos);

		// base color
		D3DXVECTOR3 c1(getRandomFloat( 0.1f,  1.0f),
					   getRandomFloat( 0.1f,  1.0f),
					   getRandomFloat( 0.1f,  1.0f));
		base_color.push_back(c1);

		// end color
		D3DXVECTOR3 c2(getRandomFloat( 0.1f,  1.0f),
					   getRandomFloat( 0.1f,  1.0f),
					   getRandomFloat( 0.1f,  1.0f));
		end_color.push_back(c2);

		// ttl, life, size
		time_to_live.push_back(5.0f);
		life.push_back(0.0f);
		size.push_back(0.5f);
	}
		
	// 3 particles (bezier curve control points) for each beam
	// point 0 is position of system
	for(i=0; i<3*num_beams; i++)
	{
		beam_particle p(getRandomFloat( 0.0f,  4.0f),	// pos x
						getRandomFloat( 0.0f,  4.0f),	// pos y
						getRandomFloat( 0.0f,  4.0f),	// pos z
						getRandomFloat( 0.0f,  4.0f),	// vx
						getRandomFloat( 0.0f,  4.0f),	// vy
						getRandomFloat( 0.0f,  4.0f));  // vz

		particle_vector.push_back(p);
	}
}

//-------------------------------------------------------------
//  beam_system::~particle_system() - destructor
//-------------------------------------------------------------
beam_system::~beam_system()
{}

//-------------------------------------------------------------
//  beam_system::update()
//-------------------------------------------------------------
void beam_system::update(float time_elapsed, int index)
{
	int i, j;
	D3DXVECTOR3 dtt;

	for(j=0; j<num_beams; j++)
	{
		life[j] += time_elapsed;

		// check for respawn
		if(life[j]>time_to_live[j])
		{
			// reset life timer
			life[j] = 0.0f; 

			// respawn the system
			for(i=j*3; i<(j+1)*3; i++)
			{
				// get random velocity
				particle_vector[i].velocity.x = -1.0f + getRandomFloat( 0.0f,  2.0f); // vel x
				particle_vector[i].velocity.y = -1.0f + getRandomFloat( 0.0f,  2.0f); // vel y
				particle_vector[i].velocity.z = -1.0f + getRandomFloat( 0.0f,  2.0f); // vel z

				// apply proper generation shape
				switch(GENERATION_SHAPE)
				{
					// tight cloud
					case 0:	particle_vector[i].position.x= -2.0f + getRandomFloat( 0.0f, 4.0f);  // pos x
							particle_vector[i].position.y= -2.0f + getRandomFloat( 0.0f, 4.0f);  // pos y
							particle_vector[i].position.z= -2.0f + getRandomFloat( 0.0f, 4.0f);  // pos z 
							break;

					// point
					case 1:	particle_vector[i].position.x= -0.2f + getRandomFloat( 0.0f, 0.4f);  // pos x
							particle_vector[i].position.y= -0.2f + getRandomFloat( 0.0f, 0.4f);  // pos y
							particle_vector[i].position.z= -0.2f + getRandomFloat( 0.0f, 0.4f);  // pos z 
							break;

					// line
					case 2:	particle_vector[i].position.x= system_position.x;  // pos x
							particle_vector[i].position.y= system_position.y;  // pos y
							particle_vector[i].position.z= -3.0f + getRandomFloat( 0.0f, 6.0f);  // pos z 
							break;

					// circle
					case 3:	particle_vector[i].position.x= -3.0f + getRandomFloat( 0.0f, 6.0f);  // pos x
							particle_vector[i].position.y= system_position.y;							 // pos y
							particle_vector[i].position.z= -3.0f + getRandomFloat( 0.0f, 6.0f);  // pos z 
							D3DXVec3Normalize(&particle_vector[i].position, &particle_vector[i].position);
							particle_vector[i].position*=3.0f;
							break;
					
					// wide cloud
					case 4:	particle_vector[i].position.x= -4.0f + getRandomFloat( 0.0f, 8.0f);  // pos x
							particle_vector[i].position.y= -4.0f + getRandomFloat( 0.0f, 8.0f);  // pos y
							particle_vector[i].position.z= -4.0f + getRandomFloat( 0.0f, 8.0f);  // pos z 
							break;
					
					// field
					case 5:	particle_vector[i].position.x= -10.0f + getRandomFloat( 0.0f, 20.0f);  // pos x
							particle_vector[i].position.y= -10.0f + getRandomFloat( 0.0f, 20.0f);  // pos y
							particle_vector[i].position.z= -10.0f + getRandomFloat( 0.0f, 20.0f);  // pos z 
				}
			}
		}

		// position updates
		for(i=j*3; i<(j+1)*3; i++)
		{
			// plug control point postion into neural net
			CPPN_VECTOR[index].nodes[0].data = particle_vector[i].position.x; 
			CPPN_VECTOR[index].nodes[1].data = particle_vector[i].position.y; 
			CPPN_VECTOR[index].nodes[2].data = particle_vector[i].position.z; 

			// plug int control point DFC
			CPPN_VECTOR[index].nodes[3].data = D3DXVec3Length(&particle_vector[i].position);

			// activate neural net
			CPPN_VECTOR[index].activate();

			// get velocity info from neural net output
			particle_vector[i].velocity.x = CPPN_VECTOR[index].nodes[7].data;
			particle_vector[i].velocity.y = CPPN_VECTOR[index].nodes[8].data;
			particle_vector[i].velocity.z = CPPN_VECTOR[index].nodes[9].data;

			// newPos = oldPos + (V*t + 1/2 A t^2)
			particle_vector[i].position += particle_vector[i].velocity * time_elapsed * ACCELERATION;
		}

		// update base color
		CPPN_VECTOR[index].nodes[0].data = 0.0f; 
		CPPN_VECTOR[index].nodes[1].data = 0.0f; 
		CPPN_VECTOR[index].nodes[2].data = 0.0f; 
		CPPN_VECTOR[index].nodes[3].data = 0.0f;

		// activate neural net
		CPPN_VECTOR[index].activate();
		
		// get color info from neural net output
		base_color[j].x = abs(CPPN_VECTOR[index].nodes[4].data);
		base_color[j].y = abs(CPPN_VECTOR[index].nodes[5].data);
		base_color[j].z = abs(CPPN_VECTOR[index].nodes[6].data);

		// update end color based on last particle
		CPPN_VECTOR[index].nodes[0].data = particle_vector[2].position.x; 
		CPPN_VECTOR[index].nodes[1].data = particle_vector[2].position.y;
		CPPN_VECTOR[index].nodes[2].data = particle_vector[2].position.z;
		CPPN_VECTOR[index].nodes[3].data = D3DXVec3Length(&particle_vector[2].position);

		// activate neural net
		CPPN_VECTOR[index].activate();
		
		// get color info from neural net output
		end_color[j].x = abs(CPPN_VECTOR[index].nodes[4].data);
		end_color[j].y = abs(CPPN_VECTOR[index].nodes[5].data);
		end_color[j].z = abs(CPPN_VECTOR[index].nodes[6].data);
	}
}

//-------------------------------------------------------------
//  beam_system::draw()
//-------------------------------------------------------------
void beam_system::draw()
{
	int i, j;
	float Ax, Ay, Az, 
		  Bx, By, Bz, 
		  Cx, Cy, Cz, 
		  Dx, Dy, Dz;

	for(j=0; j<num_beams; j++)
	{
		Ax = position[j].x;
		Ay = position[j].y;
		Az = position[j].z;
		Bx = particle_vector[j*3].position.x; 
		By = particle_vector[j*3].position.y; 
		Bz = particle_vector[j*3].position.z;
		Cx = particle_vector[j*3+1].position.x; 
		Cy = particle_vector[j*3+1].position.y; 
		Cz = particle_vector[j*3+1].position.z;
		Dx = particle_vector[j*3+2].position.x; 
		Dy = particle_vector[j*3+2].position.y; 
		Dz = particle_vector[j*3+2].position.z;

		// Points on the curve
		float X1, Y1, Z1;
		float X2, Y2, Z2;

		// Variables
		float a1 = 1.0f;
		float b1 = 1.0f - a1;

		float a2 = 1.0f - 0.1f;
		float b2 = 1.0f - a2;

		// make 20 segments of the curve 
		for(i = 0; i < 10; i++)
		{
			// Get point1 on the curve
			X1 = Ax*a1*a1*a1 + Bx*3*a1*a1*b1 + Cx*3*a1*b1*b1 + Dx*b1*b1*b1;
			Y1 = Ay*a1*a1*a1 + By*3*a1*a1*b1 + Cy*3*a1*b1*b1 + Dy*b1*b1*b1;
			Z1 = Az*a1*a1*a1 + Bz*3*a1*a1*b1 + Cz*3*a1*b1*b1 + Dz*b1*b1*b1;

			// Get a point on the curve
			X2 = Ax*a2*a2*a2 + Bx*3*a2*a2*b2 + Cx*3*a2*b2*b2 + Dx*b2*b2*b2;
			Y2 = Ay*a2*a2*a2 + By*3*a2*a2*b2 + Cy*3*a2*b2*b2 + Dy*b2*b2*b2;
			Z2 = Az*a2*a2*a2 + Bz*3*a2*a2*b2 + Cz*3*a2*b2*b2 + Dz*b2*b2*b2;
			  
			// Draw the line from point to point
			glBegin(GL_TRIANGLE_STRIP);
				glColor3fv(base_color[j]*a2 + end_color[j]*b2);
				glTexCoord2d(1,1);		
				glVertex3f(X2,Y2-size[j],Z2);
				glTexCoord2d(1,0); 
				glVertex3f(X2,Y2+size[j],Z2);

				glColor3fv(base_color[j]*a1 + end_color[j]*b1);
				glTexCoord2d(0,1); 
				glVertex3f(X1,Y1-size[j],Z1);
				glTexCoord2d(0,0); 
				glVertex3f(X1,Y1+size[j],Z1);
			glEnd();

			// increment variables
			a1 -= 0.1f;
			b1 = 1.0 - a1;

			a2 -= 0.1f;
			b2 = 1.0 - a2;
		}
	}
}

//-------------------------------------------------------------
// beam_system::reset() - resets to simple system
//-------------------------------------------------------------
void beam_system::reset(int index)
{
	// reset CPPN
	CPPN_VECTOR[index].reset(4,6);
}

#endif