//---------------------------------------------------------------------
// CPPN.h - Compositional Pattern-Producing Network
//		  - a type of Artificial Neural Network
//		  - Erin Hastings ejhastings@gmail.com
//
//		  - CPPN is composed of Nodes and Connections
//		  - input Nodes take input values
//		  - the CPPN is then activated()
//		  - output Nodes hold results of CPPN activation
//---------------------------------------------------------------------
#ifndef CPPN_H
#define CPPN_H

//---------------------------------------------------------------------
// standard includes
//---------------------------------------------------------------------
#include<fstream>
#include<iostream>
#include<vector>

//---------------------------------------------------------------------
// custom includes
//---------------------------------------------------------------------
#include "CPPNConnection.h"
#include "CPPNNode.h"
#include "Random.h"

//---------------------------------------------------------------------
// CPPN - a group of connected Nodes
//	   - input Nodes take input
//	   - data is processed through inner Nodes
//	   - output Nodes hold output 
//---------------------------------------------------------------------
class CPPN
{	
	public:	
		
	// variables				
	int generation;	
	int total_inputs;
	int total_outputs;

	// node and connection containers
	vector<Node> nodes;
	vector<Connection> connections;	

	// constructors, destructors
	CPPN();									
	~CPPN();											

	// add explicit nodes, connections
	void add_node(char TYPE);
	void add_node(char TYPE, int function);
	void add_connection(int FROM, int TO);
	void add_connection(int FROM, int TO, float WEIGHT);

	// add random nodes, connections, weights
	void add_random_connection();	
	void add_random_node();					
	void mutate_random_weight(float range); 				

	// CPPN operations
	void activate();				
	void clear();	
	void complexify();	
	bool connection_exists(int from_node, int to_node);
	void copy(CPPN &target);
	void draw(float top, float left, float width);
	void evolve_from(CPPN &target);
	void flush();						
	void reset(int num_inputs, int num_outputs);

	// console printing
	void print_connections();
	void print_nodes();

	// file reading, writing
	bool save_to_file(string filename);
	bool load_from_file(string filename);
};

//---------------------------------------------------------------------
// CPPN() - constructor
//---------------------------------------------------------------------
CPPN::CPPN()
{
	generation=0;
	total_inputs=0;
	total_outputs=0;
	nodes.reserve(25);
	connections.reserve(25);
}

//---------------------------------------------------------------------
// ~CPPN() - destructor
//---------------------------------------------------------------------
CPPN::~CPPN()
{ }

//---------------------------------------------------------------------
// add_connection() - add explicit connection from, to
//---------------------------------------------------------------------
void CPPN::add_connection(int FROM, int TO)
{
	// create new connection with weight from -1 to 1
	Connection c;
	c.from=FROM;
	c.to=TO;
	c.weight= -1.0f + getRandomFloat(0.0f, 2.0f);
	connections.push_back(c);
}

//---------------------------------------------------------------------
// add_connection() - add explicit connection from, to, weight
//---------------------------------------------------------------------
void CPPN::add_connection(int FROM, int TO, float WEIGHT)
{
	// create new connection
	Connection c;
	c.from=FROM;
	c.to=TO;
	c.weight=WEIGHT;
	connections.push_back(c);
}

//---------------------------------------------------------------------
// add_node() - add an explicit node type
//---------------------------------------------------------------------
void CPPN::add_node(char TYPE)
{
	// create new node and add to node list
	Node n;
	nodes.push_back(n);

	// 'i'==input, 'o'==output
	if(TYPE=='i') total_inputs++;
	else if(TYPE=='o') total_outputs++;
}

//---------------------------------------------------------------------
// add_node() - add an explicit node type, function
//---------------------------------------------------------------------
void CPPN::add_node(char TYPE, int FUNCTION)
{
	// create new node and add to node list
	Node n;
	n.assign_activation_function(FUNCTION);
	nodes.push_back(n);

	// 'i'==input, 'o'==output
	if(TYPE=='i') total_inputs++;
	else if(TYPE=='o') total_outputs++;
}

//---------------------------------------------------------------------
// addRandomConnection() - adds random connection
//						 - choose 2 nodes: from_node, to_node
//						 - if no connection exists add one
//
//						 - to enforce CPPN feed-forward
//						 - new connections must be from non-outputs
//						 - new connections must be to non-inputs
//---------------------------------------------------------------------
void CPPN::add_random_connection()
{
	// from_node is any non-output node
	int range = nodes.size() - total_outputs - 1;
	int from_node = getRandomInt(0, range);
	if(from_node>=total_inputs)
	{
		from_node = getRandomInt(total_inputs + total_outputs, nodes.size()-1);
	}
	
	// to node is any non-input node
	int to_node = getRandomInt(total_inputs, nodes.size()-1);

	// if to!=from && connection does not exist, add it
	if(from_node!=to_node && !connection_exists(from_node, to_node))
	{
		add_connection(from_node, to_node);
	}
}

//---------------------------------------------------------------------
// add_random_node() - add a random node + 2 connections
//					 - choose 2 nodes: from_node, to_node
//					 - adds a node and 2 connections between them
//
//					 - to enforce CPPN feed-forward
//					 - new connections must be from non-outputs
//					 - new connections must be to non-inputs
//---------------------------------------------------------------------
void CPPN::add_random_node()
{
	// from_node is any non-output node
	int range = nodes.size() - total_outputs - 1;
	int from_node = getRandomInt(0, range);
	if(from_node>=total_inputs)
	{
		from_node = getRandomInt(total_inputs + total_outputs, nodes.size()-1);
	}
	
	// to node is any non-input node
	int to_node = getRandomInt(total_inputs, nodes.size()-1);

	// find index of new node and add to node list
	int new_node = nodes.size();
	add_node('n');

	// add connnection from_node to new_node
	add_connection(from_node, new_node);

	// add connnection new_node to to_node
	add_connection(new_node, to_node);
}

//---------------------------------------------------------------------
// mutate_random_weight() - mutate connection weight by +/- range
//---------------------------------------------------------------------
void CPPN::mutate_random_weight(float range)
{
	int id = getRandomInt(0, connections.size()-1);
	connections[id].weight += -range + getRandomFloat(0.0f, range);			
}

//---------------------------------------------------------------------
// activate() - single step CPPN activation	
//			  - nodes order: |--inputs--|--outputs--|-----inner-------|
//---------------------------------------------------------------------
void CPPN::activate()
{
	int i, from, to;
	
	// reset accumulators of inner and output nodes 
	for(i=total_inputs; i<nodes.size(); i++)
	{
		nodes[i].accumulator = 0.0f;
	}

	// activate all connections 
	for(i=0; i<connections.size(); i++)
	{
		to=connections[i].to;
		from=connections[i].from;
		nodes[to].accumulator += nodes[from].data * connections[i].weight;
	}

	// apply activation function to inner and output nodes
	// save to result to data
	for(i=total_inputs; i<nodes.size(); i++)
	{
		// function pointer to activation function
		nodes[i].data = (*nodes[i].function_ptr)(nodes[i].accumulator);
	}
}

//---------------------------------------------------------------------
// CPPN::clear() - erase all nodes and connections
//---------------------------------------------------------------------
void CPPN::clear()
{
	generation=0;
	total_inputs=0;
	total_outputs=0;
	nodes.clear();
	connections.clear();						
}

//---------------------------------------------------------------------
// complexify() - chance to add node, connection, or change weight
//---------------------------------------------------------------------
void CPPN::complexify()
{
	int i, r;

	// increment generation
	generation++;

	// mutate some weights
	r = getRandomInt(1,connections.size());
	for(i=0; i<r; i++) mutate_random_weight(0.5f);

	// 30% chance to add connection
	r = getRandomInt(0,100);
	if(r<30) add_random_connection();

	// 20% chance to add node
	r = getRandomInt(0,100);
	if(r<20) add_random_node();
}

//-------------------------------------------------------------
// copy() - copy target CPPN
//-------------------------------------------------------------
void CPPN::copy(CPPN &target)
{
	int i;

	// clear existing data
	clear();

	// copy variables
	generation=target.generation;
	total_inputs=target.total_inputs;
	total_outputs=target.total_outputs;

	// copy nodes
	for(i=0; i<target.nodes.size(); i++)
	{
		Node n;
		n.assign_activation_function(target.nodes[i].activation_function);
		nodes.push_back(n);
	}

	// copy connections
	for(i=0; i<target.connections.size(); i++)
	{
		Connection c;
		c.from = target.connections[i].from;
		c.to = target.connections[i].to;
		c.weight = target.connections[i].weight;
		connections.push_back(c);
	}
}

//-------------------------------------------------------------
// connection_exists() - return true if connection exists
//-------------------------------------------------------------
bool CPPN::connection_exists(int from_node, int to_node)
{
	int i;

	for(i=0; i<connections.size(); i++)
	{
		if(connections[i].from == from_node &&
		   connections[i].to == to_node)
		{
			return true;
		}
	}

	return false;
}

//-------------------------------------------------------------
// evolve_from() - evolve from target CPPN
//				 - copy targt
//				 - chance to mutate
//-------------------------------------------------------------
void CPPN::evolve_from(CPPN &target)
{
	// copy target
	copy(target);

	// mutate
	complexify();
}

//---------------------------------------------------------------------
// flush() - set all node data to zero
//---------------------------------------------------------------------
void CPPN::flush()
{
	int i;

	for(i=0;nodes.size(); i++)
	{
		nodes[i].data=0.0f;
	}
}

//---------------------------------------------------------------------
// draw() - visual representation of a net
//		  - diagram inside a square
//---------------------------------------------------------------------
void CPPN::draw(float top, float left, float width)
{
	int i, from, to;

	// larger than normal point size
	glPointSize(8);

	// offsets between nodes
	float input_offset = width/total_inputs;
	float output_offset = width/total_outputs;
	float inner_offset = width/(nodes.size()-total_inputs-total_outputs);

	// points of square
	float mid = top-width/2.0f;
	float bottom = top-width;

	// draw connections
	glColor3f(1.0f, 1.0, 0.0f);
	glBegin(GL_LINES);

		for(i=0; i<connections.size(); i++)
		{
			// find connection from and to nodes
			from = connections[i].from;
			to = connections[i].to;

			// from is input node
			if(from<total_inputs) glVertex3f(left + input_offset * from,
											 top,
											 0.0f);

			// from is inner node
			else glVertex3f(left + inner_offset * (from-total_inputs-total_outputs),
						    mid,
						    0.0f);

			// to is output node
			if(to<total_inputs+total_outputs) glVertex3f(left + output_offset * (to - total_inputs),
														 bottom,
														 0.0f);
			// to is innner node
			else glVertex3f(left + inner_offset*(to-total_inputs-total_outputs),
							mid,
							0.0f);
		}

	glEnd();

	// draw nodes
	glBegin(GL_POINTS);

		// draw input nodes red
		// start from from top-left
		glColor3f(1.0f, 0.0, 0.0f);
		for(i=0; i<total_inputs; i++)
		{
			glVertex3f(left + input_offset*i,
					   top,
					   0.0f);
		}

		// draw inner nodes blue
		// start from bottom left
		// offset every other one
		glColor3f(0.0f, 0.0, 1.0f);
		for(i=0; i<nodes.size()-total_inputs-total_outputs; i++)
		{
			glVertex3f(left + inner_offset*i,
					   mid,
					   0.0f);
		}

		// draw output nodes green
		// start from bottom left
		glColor3f(0.0f, 1.0, 0.0f);
		for(i=0; i<total_outputs; i++)
		{
			glVertex3f(left + output_offset*i,
					   bottom,
					   0.0f);
		}

	glEnd();
}

//-------------------------------------------------------------
// reset() - resets with specified inputs/outputs
//		   - generates random inner Nodes
//-------------------------------------------------------------
void CPPN::reset(int num_inputs, int num_outputs)
{
	int i, r;

	// clear 
	clear();

	// add inputs
	for(i=0; i<num_inputs; i++) add_node('i');

	// add outputs
	for(i=0; i<num_outputs; i++) add_node('o');

	// add random nodes
	r = getRandomInt(0, 4);
	for(i=0; i<r; i++) add_random_node();

	// add random connections based on how many total nodes
	r = getRandomInt(2, (num_inputs + num_outputs) / 2);
	for(i=0; i<r; i++) add_random_connection();
}

//---------------------------------------------------------------------
// print_nodes() - print CPPN nodes to console
//---------------------------------------------------------------------
void CPPN::print_nodes()
{
	int i;
	string function_string, type_string;

	// print to console
	cout.precision(4);
	cout << endl;
	for(i=0; i<nodes.size(); i++)
	{
		// determine node function 
		switch(nodes[i].activation_function)
		{
			case 1:	function_string = "sin";
					break;

			case 2:	function_string = "cos";
					break;

			case 3:	function_string = "hyperbolic";
					break;

			case 4:	function_string = "inverse";
					break;

			case 5:	function_string = "ramp";
					break;

			case 6:	function_string = "step";
					break;

			case 7:	function_string = "gaussian";
					break;

			case 8:	function_string = "spike";
					break;

			case 9:	function_string = "tangent";
					break;

			default:function_string = "default";
		}

		// determine node type
		if(i<total_inputs) type_string = "input";
		else if(i<total_inputs+total_outputs) type_string="output";
		else type_string = "inner";

		// print to console
		cout << "Node:" << i << " ";
		cout << " Type:" << type_string;
		cout <<	" Function:" << function_string;
		cout << " Data:" << nodes[i].data;
		cout << " Accumulator:" << nodes[i].accumulator;
		cout << endl;
	}
	cout << endl;
}

//---------------------------------------------------------------------
// print_connections() - print CPPN connections to console
//---------------------------------------------------------------------
void CPPN::print_connections()
{
	int i;

	// print to console
	cout.precision(4);
	cout << endl;
	for(i=0; i<connections.size(); i++ )
	{
		cout << "Connection:" << i;
		cout << " From:" << connections[i].from;
		cout << " To:" << connections[i].to;
		cout << " Weight:" << connections[i].weight;
		cout << endl;
	}
   	cout << endl;
}

//---------------------------------------------------------------------
// load_from_file(() - load CPPN from file
//---------------------------------------------------------------------
bool CPPN::load_from_file(string filename)
{
	// open the file
	ifstream inputFile(filename.c_str());
	if(!inputFile) return false;

	// clear existing CPPN
	clear();

	string S;
	int function;

	// read input file into the CPPN
	while (inputFile>>S)	
	{
		if(S=="generation") 
		{
			inputFile >> generation;
		}
		else if(S=="inputs") 
		{
			inputFile >> total_inputs;
		}
		else if(S=="outputs") 
		{
			inputFile >> total_outputs;
		}
		else if(S=="node")
		{
			inputFile >> function;
			Node n;
			n.assign_activation_function(function);
			nodes.push_back(n);
		}

		else if(S=="connection")
		{
			Connection c;
			inputFile >> c.from;
			inputFile >> c.to;
			inputFile >> c.weight;
			connections.push_back(c);
		}
		else
		{
			// file error
			return false;
		}
	}

	// close file
	inputFile.close();
	return true;
}

//---------------------------------------------------------------------
// save_to_file() - save CPPN to file
//---------------------------------------------------------------------
bool CPPN::save_to_file(string filename)
{
	int i;
	ofstream outputFile(filename.c_str());

	// check if error opening file
	if(!outputFile) return false;

	// variables
	outputFile << "generation " << generation << endl;
	outputFile << "inputs " << total_inputs << endl;
	outputFile << "outputs " << total_outputs << endl;

	// write nodes
	for(i=0; i<nodes.size(); i++)
	{
		outputFile << "node " 
			       << nodes[i].activation_function
				   << endl;
	}

	// write connections
	for(i=0; i<connections.size(); i++)
	{
		outputFile << "connection " 
			       << connections[i].from 
				   << " " 
				   << connections[i].to 
				   << " "
				   << connections[i].weight
				   << endl;
	}

	// flush file stream and close
	outputFile.flush();
	outputFile.close();
	return true;
}

#endif