//---------------------------------------------------------------------
// Keys.cpp - handles keyboard input
//			- some logic occurs here after each keypress
//---------------------------------------------------------------------
#include <d3dx9.h>					// DirectX 9
#include <gl/gl.h>					// OpenGL32 Library
#include <gl/glu.h>					// GLu32 Library
#include <string>					// string class
#include <windows.h>				// Windows Library
#include "NeHeGL.h"					// Header File For NeHeGL
using namespace std;

//---------------------------------------------------------------------
// external variables
//---------------------------------------------------------------------
extern float KEY_COOLDOWN;			// global cooldown on keys
extern D3DXVECTOR3 THETA;			// rotation of the axis
extern bool ZOOM;					// zoom in?
extern bool PAUSE;					// pause animation?
extern int SELECTED;				// selected system
extern int ACTIVATION_FUNCTION;		// activation function
extern string FUNCTION_STRING;		// name of activation function
extern int GENERATION_SHAPE;		// generation shape
extern string SHAPE_STRING;			// name of generation shape	

extern void reset_all();			// reset function
extern void save_all();
extern void load_all();				// load CPPNs
extern void complexify(int id);		// complexify function
extern void lower_density();		// lower density
extern void raise_density();		// raise density
extern void new_generation(int id);	// new generation based on selected system
extern void undo();					// undo function
extern void copy_to_undo_vector();  // copy CPPN vector to UNDO_VECTOR
extern void load_CPPN(int index);	// load a CPPN
extern void save_CPPN(int index);	// save a CPPN
extern void switch_textures(int id);// switches textures

extern float ACCELERATION;			// speed of particle animation
extern int CURRENT_TEXTURE;			// current texture
extern int PARTICLE_CLASS;			// class of particle system
extern string CLASS_STRING;			// name of particle class
extern int PARTICLE_DENSITY;		// amount of particles
extern bool TRAIL_LINES;			// draw trail lines?

//---------------------------------------------------------------------
// KeyProcess - handles keyboard input
//---------------------------------------------------------------------
void keyProcess(float time_elapsed)
{
	// arrows keys = rotate view
	// time_elapsed used to make turning rate same
	// regardless of frame rate
	if (g_keys->keyDown[VK_RIGHT])					
	{
		THETA.y += 100.0f * time_elapsed;			
	}	
	if (g_keys->keyDown[VK_LEFT])							
	{
		THETA.y -= 100.0f * time_elapsed;		
	}
	if (g_keys->keyDown[VK_UP])							
	{
		THETA.x += 100.0f * time_elapsed;				
	}
	if (g_keys->keyDown[VK_DOWN])							
	{
		THETA.x -= 100.0f * time_elapsed;
	}

	// check key cooldown timer
	// prevents over-sensitive keys
	if(KEY_COOLDOWN > 200.0f)	
	{
		// ESC = exit program
		if (g_keys->keyDown[VK_ESCAPE])			
		{
			TerminateApplication (g_window);			
		}

		// 'Tab' select
		if (g_keys->keyDown[VK_TAB])					
		{
			SELECTED++;
			if(SELECTED>8) SELECTED=0;
			PlaySound("sounds/select.wav", NULL, SND_ASYNC); 	
		}

		// 'C' change particle class
		//	load saved CPPN from load files
		else if (g_keys->keyDown['C'])					
		{
			PARTICLE_CLASS++;
			if(PARTICLE_CLASS>4) PARTICLE_CLASS=0;

			// switch to some suggested variables as well
			switch(PARTICLE_CLASS)
			{
				case 0:	CLASS_STRING="Base";
						ACCELERATION=1;
						switch_textures(0);
						break;
				case 1:	CLASS_STRING="Beam";
						ACCELERATION=3;
						switch_textures(12);
						break;
				case 2:	CLASS_STRING="Sheet"; 
						ACCELERATION=3;
						switch_textures(5);
						break;
				case 3:	CLASS_STRING="Rotator"; 
						ACCELERATION=5;
						switch_textures(0);
						break;
				case 4:	CLASS_STRING="Trail"; 
						ACCELERATION=1;
						switch_textures(0);
						break;
			}

			// load saved CPPNs for this class
			load_all();

			// play sound and zoom out
			PlaySound("sounds/select.wav", NULL, SND_ASYNC);
			ZOOM=false;
		}

		// 'L' load  CPPN from file
		else if (g_keys->keyDown['L'])						
		{
			load_CPPN(SELECTED);
			PlaySound("sounds/save.wav", NULL, SND_ASYNC); 
		}

		// 'G' generation shape 
		else if (g_keys->keyDown['G'])					
		{
			GENERATION_SHAPE++;
			if(GENERATION_SHAPE>5) GENERATION_SHAPE=0;
			switch(GENERATION_SHAPE)
			{
				case 0:	SHAPE_STRING="TightCloud"; 
						break;
				case 1:	SHAPE_STRING="Point"; 
						break;
				case 2:	SHAPE_STRING="Line"; 
						break;
				case 3:	SHAPE_STRING="Circle"; 
						break;
				case 4:	SHAPE_STRING="WideCloud"; 
						break;
				case 5:	SHAPE_STRING="Field"; 
						break;
			}
			PlaySound("sounds/select.wav", NULL, SND_ASYNC); 
		}

		// 'N' new generation
		else if (g_keys->keyDown['N'])					
		{
			// first backup existing CPPN_VECTOR
			copy_to_undo_vector();

			// create new generation based on selected CPPN
			new_generation(SELECTED);

			// zoom out and play sound
			if(ZOOM) ZOOM=false;
			PlaySound("sounds/zoom.wav", NULL, SND_ASYNC); 
		}

		// 'P' pause
		else if (g_keys->keyDown['P'])			
		{
			if(PAUSE) PAUSE=false;
			else PAUSE=true;
			PlaySound("sounds/mode.wav", NULL, SND_ASYNC); 
		}

		// 'R' reset particle evolution
		else if (g_keys->keyDown['R'])			
		{
			reset_all();
			ZOOM=false;
			PlaySound("sounds/mode.wav", NULL, SND_ASYNC); 
		}
		
		// 'S' speed of particle animation
		else if (g_keys->keyDown['S'])				
		{
			ACCELERATION+=1.0f;
			if(ACCELERATION>10.0f) ACCELERATION=0.0f;
			PlaySound("sounds/select.wav", NULL, SND_ASYNC); 		
		}

		// 'T' texture swap
		else if (g_keys->keyDown['T'])						
		{
			switch_textures(-1);
			PlaySound("sounds/mode.wav", NULL, SND_ASYNC); 
		}

		// 'U' undo	
		else if (g_keys->keyDown['U'])					
		{
			undo();
			PlaySound("sounds/save.wav", NULL, SND_ASYNC); 
		}

		// 'V' save CPPN to file
		else if (g_keys->keyDown['V'])						
		{
			save_CPPN(SELECTED);
			PlaySound("sounds/save.wav", NULL, SND_ASYNC); 
		}
		
		// 'Z' zoom	
		else if (g_keys->keyDown['Z'])					
		{
			if(ZOOM) ZOOM=false;
			else ZOOM=true;		
			PlaySound("sounds/zoom.wav", NULL, SND_ASYNC); 
		}

		// '+' raise raise particle density 
		else if (g_keys->keyDown[VK_OEM_PLUS])					
		{
			PARTICLE_DENSITY+=5;
			PlaySound("sounds/select.wav", NULL, SND_ASYNC); 	
			raise_density();
		}
			
		// '-' lower particle density
		else if (g_keys->keyDown[VK_OEM_MINUS])					
		{
			PARTICLE_DENSITY-=5;
			if(PARTICLE_DENSITY<1) PARTICLE_DENSITY=1;
			PlaySound("sounds/select.wav", NULL, SND_ASYNC);
			lower_density();
		}

		// reset key cooldown timer
		KEY_COOLDOWN = 0.0f;
	}
}