//---------------------------------------------------------------------
// Main.cpp - NEAT Particles by Erin Hastings
//			- email: hastings@cs.ucf.edu
//---------------------------------------------------------------------
#define _CRT_SECURE_NO_DEPRECATE 1	// prevent annoying compiler warnings
#pragma warning(disable : 4018)		// signed/unsigned mismatch
#pragma warning(disable : 4267)		// conversion from 'size_t' to 'int'
#pragma warning(disable : 4244)		// conversion from 'time_t' to 'unsigned int'
#pragma warning(disable : 4312)		// conversion from 'LONG' to 'GL_Window *' of greater size
#pragma warning(disable : 4311)		// pointer truncation from 'GL_Window *' to 'LONG'

//---------------------------------------------------------------------
// standard includes
//---------------------------------------------------------------------
#include <windows.h>				// For Windows
#include <d3dx9.h>					// DirectX 9
#include <gl\gl.h>					// OpenGL32 Library
#include <gl\glu.h>					// GLu32 Library
#include <gl\glaux.h>				// GLaux Library
#include <math.h>					// Math Operations
#include <stdio.h>					// Standard Input / Output
#include <stdarg.h>					// Variable Argument Routines
#include <string>					// string class
#include <sstream>					// string stream
#include <time.h>					// for timing and random number seeding
#include <vector>					// stl vector

//---------------------------------------------------------------------
// standard libraries
//---------------------------------------------------------------------
#pragma comment(lib,"d3d9.lib")		// D3D9.lib
#pragma comment(lib,"d3dx9.lib")	// D3DX9.lib
#pragma comment(lib,"glu32.lib")	// GLu32.lib 
#pragma comment(lib,"glaux.lib")	// GLaux.lib 
#pragma comment(lib,"opengl32.lib")	// OpenGL32.lib 
#pragma comment(lib,"winmm.lib")	// Windows multimedia library

//---------------------------------------------------------------------
// custom includes
//---------------------------------------------------------------------
#include "Random.h"					// random number generation functions
#include "CPPN.h"					// neural net class
#include "Billboard.h"				// sets up particle billboards
#include "Font.h"					// font support
#include "Lines.h"					// gl line drawing functions
#include "NeHeGL.h"					// Header File For NeHeGL
#include "ParticleSystem.h"			// particle system class
#include "BeamSystem.h"				// beam particle class
#include "PlaneSystem.h"			// plane particle class
#include "RotatorSystem.h"			// rotator particle system
#include "TrailSystem.h"			// trail particle system
#include "Texture.h"				// texture loading / storage functions
#include "FileIO.h"					// file reading / writing 

//---------------------------------------------------------------------
// other compiler stuff
//---------------------------------------------------------------------
using namespace std;				// use c++ stl

//---------------------------------------------------------------------
// GL framework stuff
//---------------------------------------------------------------------
#ifndef		CDS_FULLSCREEN			// CDS_FULLSCREEN Is Not Defined By Some
#define		CDS_FULLSCREEN 4		// Compilers. By Defining It This Way,
#endif								// We Can Avoid Errors

GL_Window*	g_window;
Keys*		g_keys;

char	appTitle[]="NEAT Particles";		//stores the window title
int		screenInfo[3]={1024,768,32};		//stores resolution and color bits (w,h,bpp)

//---------------------------------------------------------------------
// function declarations
//---------------------------------------------------------------------
BOOL Initialize (GL_Window* window, Keys* keys);
void Deinitialize (void);	
void Update(DWORD milliseconds);
void DrawScene(void);
void reset_all();

//---------------------------------------------------------------------
// camera related variables
//---------------------------------------------------------------------
D3DXVECTOR3 CAMERA(0.0f, 12.0f, 12.0f);		// camera position
D3DXVECTOR3 WORLD_UP(0.0f, 1.0f, 0.0f);		// up vector in scene
D3DXVECTOR3 LOOK_AT(0.0f, 0.0f, 0.0f);		// where the camera is looking
D3DXVECTOR3 THETA(0.0f, 30.0f, 0.0f);		// rotation of the world axis

//---------------------------------------------------------------------
// billboarding/particle related variables
//---------------------------------------------------------------------
int CURRENT_TEXTURE = 0;	// which texture is being used?
D3DXVECTOR3 RIGHT; 			// billboard corners
D3DXVECTOR3 UP;				// up vector for billboarding

//---------------------------------------------------------------------
// font related variable
//---------------------------------------------------------------------
GLuint FONT_BASE;			// Base Display List For The Font Set

//---------------------------------------------------------------------
// state variables
//---------------------------------------------------------------------
float KEY_COOLDOWN=0.0f;					// global cooldown on keys
int SELECTED = 0;							// which systems is selected				
bool ZOOM = false;							// zoom in?
bool PAUSE = false;							// pause updates on all but selected?
int PARTICLE_DENSITY=30;					// amount of particles
float ACCELERATION=1.0f;					// speed of particle animation
int ACTIVATION_FUNCTION=0;					// activation function
string FUNCTION_STRING = "Bipolar Sigmoid"; // name of activation function
int GENERATION_SHAPE=0;						// generation shape
string SHAPE_STRING="Sphere";				// name of generation shape
int PHYSICS_TYPE=0;							// physics model
int PARTICLE_CLASS=0;						// class of particle system
string CLASS_STRING="Generic";				// class name
int SAVE_FILE;								// next save file CPPN_X.txt 
int FPS = 0;								// fps display value
int FPS_COUNTER=0;							// fps counter
float FPS_TIMER=0.0f;						// fps timer

//---------------------------------------------------------------------
// object lists
//---------------------------------------------------------------------
vector<texture> TEXTURE_VECTOR;
vector<CPPN> CPPN_VECTOR;
vector<particle_system> PARTICLE_SYSTEM_VECTOR;
vector<beam_system> BEAM_SYSTEM_VECTOR;
vector<plane_system> PLANE_SYSTEM_VECTOR;
vector<rotator_system> ROTATOR_SYSTEM_VECTOR;
vector<trail_system> TRAIL_SYSTEM_VECTOR;
vector<CPPN> UNDO_VECTOR;

//---------------------------------------------------------------------
// Initialize() - all pre-execution setup goes here
//---------------------------------------------------------------------
BOOL Initialize (GL_Window* window, Keys* keys)			
{
	int i;

	// window and key pointers
	g_window	= window;
	g_keys		= keys;

	// gl state setup
	glClearColor (0.0f, 0.0f, 0.0f, 0.5f);				
	glShadeModel (GL_SMOOTH);							
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);	
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);								

	// build gl font list
	BuildFont(FONT_BASE);		

	// seed random number generator
	srand ( time(NULL) );

	// load textures
	ifstream textureFile;
	textureFile.open("textures/texture_list.txt");
	string S;
	while(textureFile >> S)
	{
		texture t;
		t.load(S);
		TEXTURE_VECTOR.push_back(t);
	}

	// bind default texture
	glBindTexture(GL_TEXTURE_2D, TEXTURE_VECTOR[0].data[0]);

	// create particle systems
	for(i=0; i<9; i++)
	{
		particle_system p;
		PARTICLE_SYSTEM_VECTOR.push_back(p);
	}

	// create beam systems
	for(i=0; i<9; i++)
	{
		beam_system b;
		BEAM_SYSTEM_VECTOR.push_back(b);
	}

	// create plane systems
	for(i=0; i<9; i++)
	{
		plane_system p;
		PLANE_SYSTEM_VECTOR.push_back(p);
	}

	// create rotator systems
	for(i=0; i<9; i++)
	{
		rotator_system r;
		ROTATOR_SYSTEM_VECTOR.push_back(r);
	}

	// create trail systems
	for(i=0; i<9; i++)
	{
		trail_system r;
		TRAIL_SYSTEM_VECTOR.push_back(r);
	}

	// create neural nets
	for(i=0; i<9; i++)
	{
		CPPN N;
		N.reset(4,6);
		N.complexify();
		CPPN_VECTOR.push_back(N);
	}
	
	// load saved CPPNs
	load_all();

	// initialize undo list
	for(i=0; i<9; i++)
	{
		CPPN N;
		N.copy(CPPN_VECTOR[i]);
		UNDO_VECTOR.push_back(N);
	}
	D3DXVECTOR3 V(0.0f, 0.0f, 0.0f);
	float M[3] = {0.0f, 0.0f, 0.0f};

	glVertex3fv(V);
	glVertex3fv(M);

	return TRUE;		
}

//---------------------------------------------------------------------
// switch_textures() - change to new texture
//---------------------------------------------------------------------
void switch_textures(int id)						
{
	if(id<0)
	{
		// increment current texture
		CURRENT_TEXTURE++;
		if(CURRENT_TEXTURE>=TEXTURE_VECTOR.size())
		{
			CURRENT_TEXTURE=0;
		}
		
		// bind texture from the texture list
		glBindTexture(GL_TEXTURE_2D, TEXTURE_VECTOR[CURRENT_TEXTURE].data[0]);
	}
	else glBindTexture(GL_TEXTURE_2D, TEXTURE_VECTOR[id].data[0]);
}

//---------------------------------------------------------------------
// DeInitialize() - post execution cleanup goes here
//---------------------------------------------------------------------
void Deinitialize (void)						
{
	KillFont(FONT_BASE);
}

//---------------------------------------------------------------------
// reset_all() - resets all particle systems
//---------------------------------------------------------------------
void reset_all()
{
	int i;
					
	if(PARTICLE_CLASS==0)
	{
		// reset particle systems
		for(i=0; i<PARTICLE_SYSTEM_VECTOR.size(); i++)
		{
			PARTICLE_SYSTEM_VECTOR[i].reset(i);
		}
	}

	else if(PARTICLE_CLASS==1)
	{
		// reset beam systems
		for(i=0; i<BEAM_SYSTEM_VECTOR.size(); i++)
		{
			BEAM_SYSTEM_VECTOR[i].reset(i);
		}
	}

	else if(PARTICLE_CLASS==2)
	{
		// reset plane systems
		for(i=0; i<PLANE_SYSTEM_VECTOR.size(); i++)
		{
			PLANE_SYSTEM_VECTOR[i].reset(i);
		}
	}

	else if(PARTICLE_CLASS==3)
	{
		// reset rotator systems
		for(i=0; i<ROTATOR_SYSTEM_VECTOR.size(); i++)
		{
			ROTATOR_SYSTEM_VECTOR[i].reset(i);
		}
	}

	else if(PARTICLE_CLASS==4)
	{
		// reset trail systems
		for(i=0; i<ROTATOR_SYSTEM_VECTOR.size(); i++)
		{
			TRAIL_SYSTEM_VECTOR[i].reset(i);
		}
	}
}

//---------------------------------------------------------------------
// lower_density() - changes particle density
//				   - only applies to particle/trail systems
//---------------------------------------------------------------------
void lower_density()
{
	// lower particle system density
	if(PARTICLE_SYSTEM_VECTOR[0].particle_vector.size() > 10)
	{
		int	i;
		for(i=0; i<PARTICLE_SYSTEM_VECTOR.size(); i++)
		{
			PARTICLE_SYSTEM_VECTOR[i].particle_vector.erase(PARTICLE_SYSTEM_VECTOR[i].particle_vector.end()-6,
															PARTICLE_SYSTEM_VECTOR[i].particle_vector.end()-1);
		}
	}

	// make sure density is > 10
	if(TRAIL_SYSTEM_VECTOR[0].particle_vector.size() > 10)
	{
		int	i;
		for(i=0; i<TRAIL_SYSTEM_VECTOR.size(); i++)
		{
			TRAIL_SYSTEM_VECTOR[i].particle_vector.erase(TRAIL_SYSTEM_VECTOR[i].particle_vector.end()-6,
															TRAIL_SYSTEM_VECTOR[i].particle_vector.end()-1);
		}
	}

	PARTICLE_DENSITY = TRAIL_SYSTEM_VECTOR[0].particle_vector.size();
}

//---------------------------------------------------------------------
// lower_density() - changes particle density
//				   - only applies to particle/trail systems
//---------------------------------------------------------------------
void raise_density()
{
	int	i,j;

	// generic system
	for(i=0; i<PARTICLE_SYSTEM_VECTOR.size(); i++)
	{
		for(j=0; j<5; j++)
		{
			particle p;
			PARTICLE_SYSTEM_VECTOR[i].particle_vector.push_back(p);
		}
	}

	// trail system
	for(i=0; i<TRAIL_SYSTEM_VECTOR.size(); i++)
	{
		for(j=0; j<5; j++)
		{
			trail_particle p;
			TRAIL_SYSTEM_VECTOR[i].particle_vector.push_back(p);
		}
	}

	// set particle density global
	PARTICLE_DENSITY = TRAIL_SYSTEM_VECTOR[0].particle_vector.size();
}

//---------------------------------------------------------------------
// copy_to_undo_vector()() - copy all CPPN_VECTOR to UNDO_VECTOR
//---------------------------------------------------------------------
void copy_to_undo_vector()
{
	int i;

	// copy all CPPN_VECTOR to UNDO_VECTOR
	for(i=0; i<9; i++)
	{
		UNDO_VECTOR[i].copy(CPPN_VECTOR[i]);
	}
}

//---------------------------------------------------------------------
// undo() - revert to previous generation
//		  - previous generation CPPN saved in UNDO_VECTOR
//---------------------------------------------------------------------
void undo()
{
	int i;

	// copy CPPNs from previous generation
	for(i=0; i<9; i++)
	{
		CPPN_VECTOR[i].copy(UNDO_VECTOR[i]);
	}
}

//---------------------------------------------------------------------
// new_generation() - creates a new generation based on selected CPPN
//---------------------------------------------------------------------
void new_generation(int id)
{
	int i;

	// spawn new generation based on selected CPPN
	for(i=0; i<CPPN_VECTOR.size(); i++)
	{
		// dont evolve selected CPPN
		if(i!=id) CPPN_VECTOR[i].evolve_from(CPPN_VECTOR[id]);
	}
}

//---------------------------------------------------------------------
// Update() - runs every frame of animation
//			- updates all objects in the scene
//			- based on milliseconds elapsed since previous frame
//---------------------------------------------------------------------
void Update(DWORD milliseconds)									
{
	// update key cooldown timer
	KEY_COOLDOWN+=milliseconds;	

	// convert dword to float
	float time_elapsed = milliseconds * 0.001f;

	// handle keyboard input
	keyProcess(time_elapsed);

	// update FPS
	FPS_COUNTER++;
	FPS_TIMER+=milliseconds;
	if(FPS_TIMER>1000)
	{
		FPS = FPS_COUNTER;
		FPS_COUNTER = 0;
		FPS_TIMER=0.0f;
	}

	if(ZOOM && !PAUSE)	
	{
		// zoom mode, only update selected item
		if(PARTICLE_CLASS==0) 
			PARTICLE_SYSTEM_VECTOR[SELECTED].update(time_elapsed, SELECTED);
		else if(PARTICLE_CLASS==1) 
			BEAM_SYSTEM_VECTOR[SELECTED].update(time_elapsed, SELECTED);
		else if(PARTICLE_CLASS==2) 
			PLANE_SYSTEM_VECTOR[SELECTED].update(time_elapsed, SELECTED);
		else if(PARTICLE_CLASS==3) 
			ROTATOR_SYSTEM_VECTOR[SELECTED].update(time_elapsed, SELECTED);
		else if(PARTICLE_CLASS==4) 
			TRAIL_SYSTEM_VECTOR[SELECTED].update(time_elapsed, SELECTED);
	}
	else if(!PAUSE)
	{
		// normal mode, update all 9 items
		int i;
		if(PARTICLE_CLASS==0) 
		{
			for(i=0; i<PARTICLE_SYSTEM_VECTOR.size(); i++)		
				PARTICLE_SYSTEM_VECTOR[i].update(time_elapsed, i);
		}
		else if(PARTICLE_CLASS==1) 
		{
			for(i=0; i<BEAM_SYSTEM_VECTOR.size(); i++)		
				BEAM_SYSTEM_VECTOR[i].update(time_elapsed, i);
		}
		else if(PARTICLE_CLASS==2) 
		{
			for(i=0; i<PLANE_SYSTEM_VECTOR.size(); i++)		
				PLANE_SYSTEM_VECTOR[i].update(time_elapsed, i);
		}
		else if(PARTICLE_CLASS==3) 
		{
			for(i=0; i<ROTATOR_SYSTEM_VECTOR.size(); i++)		
				ROTATOR_SYSTEM_VECTOR[i].update(time_elapsed, i);
		}
		else if(PARTICLE_CLASS==4) 
		{
			for(i=0; i<TRAIL_SYSTEM_VECTOR.size(); i++)		
				TRAIL_SYSTEM_VECTOR[i].update(time_elapsed, i);
		}
	}
}

//---------------------------------------------------------------------
// Draw() - runs every frame of animation
//		  - draws all objects in the scene
//---------------------------------------------------------------------
void DrawScene(void)											
{
	int i;
	RECT	rect;								// Holds Coordinates Of A Rectangle
	GetClientRect(g_window->hWnd, &rect);		// Get Window Dimensions
	int window_width=rect.right-rect.left;		// Calculate The Width (Right Side-Left Side)
	int window_height=rect.bottom-rect.top;		// Calculate The Height (Bottom-Top)
	glClear (GL_COLOR_BUFFER_BIT);				// Clear Screen

	// zoomed in on a single item
	if(ZOOM) 
	{
		glClear (GL_DEPTH_BUFFER_BIT);					

		glViewport (0,				// x
					0,				// y
					window_width,	// width
					window_height);	// height	
		glLoadIdentity ();

		gluLookAt(CAMERA.x+3.0f, CAMERA.y+3.0f, CAMERA.z+3.0f,				
				  LOOK_AT.x, LOOK_AT.y, LOOK_AT.z,
				  WORLD_UP.x, WORLD_UP.y, WORLD_UP.z);

		// move the object slightly left
		glPushMatrix();
		glTranslatef(-3.0f, 0.0f, 0.0f);

			// draw the corner numbers
			glColor3f(1.0f, 1.0f, 0.0f);
			glRasterPos2f(-6, 6);
			glPrint(FONT_BASE, "%d", SELECTED);	

			// rotate the scene
			glRotatef(THETA.x, 1.0, 0.0, 0.0);			
			glRotatef(THETA.y, 0.0, 1.0, 0.0);
			glRotatef(THETA.z, 0.0, 0.0, 1.0);

			// draw axis
			DrawAxis();

			// enable textures and blending 
			// for particle rendering
			glEnable(GL_TEXTURE_2D);		
			glEnable(GL_BLEND);	
			
			// draw particle systems
			if(PARTICLE_CLASS==0)
			{
				// generic
				SetupBillboards(RIGHT, UP);	
				PARTICLE_SYSTEM_VECTOR[SELECTED].draw();
			}
			else if(PARTICLE_CLASS==1)
			{
				// beam
				BEAM_SYSTEM_VECTOR[SELECTED].draw();
			}
			else if(PARTICLE_CLASS==2)
			{
				// plane
				PLANE_SYSTEM_VECTOR[SELECTED].draw();
			}
			else if(PARTICLE_CLASS==3)
			{
				// rotator
				SetupBillboards(RIGHT, UP);
				ROTATOR_SYSTEM_VECTOR[SELECTED].draw();
			}
			else if(PARTICLE_CLASS==4)
			{
				// trail
				SetupBillboards(RIGHT, UP);	
				TRAIL_SYSTEM_VECTOR[SELECTED].draw();
			}

	    // remove transforms
		glPopMatrix();

		// disable texturing and blending
		glDisable(GL_BLEND);			
		glDisable(GL_TEXTURE_2D);
	}

	// not zoomed, split into 9 gl viewports
	else 
	{
		unsigned int j, k;

		for(i=0, j=0; j<3; j++)
		{
			for(k=0; k<3; k++, i++)
			{
				glClear (GL_DEPTH_BUFFER_BIT);					

				glViewport (k*window_width/3,	// x
							window_height - 
							window_height/3 -
							j*window_height/3,	// y
							window_width/3,		// width
							window_height/3);	// height	
				glLoadIdentity ();

				gluLookAt(CAMERA.x, CAMERA.y, CAMERA.z,				
						  LOOK_AT.x, LOOK_AT.y, LOOK_AT.z,
						  WORLD_UP.x, WORLD_UP.y, WORLD_UP.z);


				// draw the corner numbers
				if(SELECTED==i) glColor3f(1.0f, 0.0f, 0.0f);
				else  glColor3f(1.0f, 1.0f, 0.0f);
				glRasterPos2f(-6, 6);
				glPrint(FONT_BASE, "%d", i);	

				// rotate the scene
				glRotatef(THETA.x, 1.0, 0.0, 0.0);			
				glRotatef(THETA.y, 0.0, 1.0, 0.0);
				glRotatef(THETA.z, 0.0, 0.0, 1.0);

				// draw axis	
				DrawAxis();

				// enable textures and blending 
				// for particle rendering
				glEnable(GL_TEXTURE_2D);		
				glEnable(GL_BLEND);	
				
				// switch on particle_class
				if(PARTICLE_CLASS==0)
				{	
					// draw particle systems
					SetupBillboards(RIGHT, UP);			
					PARTICLE_SYSTEM_VECTOR[i].draw();
				}
				else if(PARTICLE_CLASS==1)
				{
					// draw beam systems	
					BEAM_SYSTEM_VECTOR[i].draw();
				}
				else if(PARTICLE_CLASS==2)
				{
					// draw plane systems
					PLANE_SYSTEM_VECTOR[i].draw();
				}
				else if(PARTICLE_CLASS==3)
				{
					// draw rotator systems
					SetupBillboards(RIGHT, UP);	
					ROTATOR_SYSTEM_VECTOR[i].draw();
				}
				else if(PARTICLE_CLASS==4)
				{
					// draw trail systems	
					SetupBillboards(RIGHT, UP);	
					TRAIL_SYSTEM_VECTOR[i].draw();
				}

				// done with particle drawing
				// disable texturing and blending
				glDisable(GL_BLEND);			
				glDisable(GL_TEXTURE_2D);
			}
		}
	}

	// go back to fullscren viewport
	// to draw text and CPPNs
	glClear (GL_DEPTH_BUFFER_BIT);
	glViewport (0,					// x
				0,					// y
				window_width,		// width
				window_height);		// height
	glLoadIdentity ();
	gluLookAt(0, 0, 20,				// set the camera position
			  0, 0, 0,
			  0, 1, 0);

	// draw the neural nets
	if(ZOOM) 
	{
		CPPN_VECTOR[SELECTED].draw(4.0f, 4.0f, 6.0f);	
	}

	// draw the application data at screen bottom
	glColor3f(1.0f, 1.0f, 0.0f);
	glRasterPos2f(-9.5, -7.5);
	glPrint(FONT_BASE, "'C'class=%s  'G'enerationShape=%s  'T'exture=%s  'S'speed=%2.0f  '+/-'density=%d", CLASS_STRING.c_str(), 
																										    SHAPE_STRING.c_str(),
																										    TEXTURE_VECTOR[CURRENT_TEXTURE].filename.c_str(),
																										    ACCELERATION,
																										    PARTICLE_DENSITY);	
	glRasterPos2f(-9.5, -8);
	glPrint(FONT_BASE, "'Arrows'rotate  'Tab'select  'Z'oom  'P'ause  'R'eset  'C'omplexify  'N'ewGeneration  'V'Save  'L'oad  FPS=%d", FPS);
	
	// draw scene
	glFlush();											
}