//---------------------------------------------------------------------
// neural_net.cpp - neural net class defined here
//				  - a neural net is composed of nodes and connections
//				  - input nodes contain input
//				  - output nodes hold results of the network
//---------------------------------------------------------------------

#ifndef NEURAL_NET_H
#define NEURAL_NET_H
		
#include <fstream>					
#include <iostream>
#include <vector>
#include <string>
#include "connection.h"
#include "node.h"
#include "random.h"
using namespace std;

// extern variables
extern int ACTIVATION_FUNCTION;

//---------------------------------------------------------------------
// neural_net - a group of connected nodes
//			  - input nodes take input
//			  - data is processed through inner nodes
//			  - output nodes hold output 
//---------------------------------------------------------------------
class neural_net
{	
	public:	int id;									// id of net				
			string tag;								// name of net
			int generation;							// generation of evolution

			vector<node> node_vector;				// vector of all nodes
			vector<connection> connection_vector;	// vector of all connections
			vector<int> input_vector;				// indexes of all input nodes
			vector<int> output_vector;				// indexes of all output nodes

			neural_net();							// default contructor
			~neural_net();							// destructor

			neural_net(int id,
					   string tag,
					   int generation);				// explicit contructor

			void addNode(int id, 					// add a new node
						 string tag,				
						 char type,						
						 float data);				
								

			void addConnection(int id, 				// add a new connection
							   string tag,			
							   int from,			
							   int to,
							   float weight);

			void mutate_weight(int id); 		// mutate connection weight

			void activate();

			void print_attributes();			// print functions
			void print_nodes();
			void print_connections();
			void print_input_nodes();
			void print_output_nodes();

			void draw();						// draw the net in 3d
			void complexify();					// add node and/or link
};

//---------------------------------------------------------------------
// neural_net() - constructor
//---------------------------------------------------------------------
neural_net::neural_net()
{
	id=-1;
	generation=0;
	tag="net";									
}

//---------------------------------------------------------------------
// ~neural_net() - destructor
//---------------------------------------------------------------------
neural_net::~neural_net()
{ }

//---------------------------------------------------------------------
// neural_net() - constructor
//---------------------------------------------------------------------
neural_net::neural_net(int i, string t, int g)
{
	id=i;						
	tag=t;							
	generation=g;
}

//---------------------------------------------------------------------
// add_node() - mutates the network by adding a node
//---------------------------------------------------------------------
void neural_net::addNode(int id, string tag, char type, float data)
{
	node new_node(id, tag, type, data);
	if(type=='i') input_vector.push_back(node_vector.size());
	else if(type=='o') output_vector.push_back(node_vector.size());
	node_vector.push_back( new_node );
}

//---------------------------------------------------------------------
// add_connection() - mutates the network by adding a connection
//---------------------------------------------------------------------
void neural_net::addConnection(int i, string tg, int f, int t, float w)
{
	connection new_connection(i, tg, f, t, w);
	connection_vector.push_back( new_connection );
}

//---------------------------------------------------------------------
// mutate_weight() - mutates connection weight by a random number
//				   - from -1.0 to 1.0
//---------------------------------------------------------------------
void neural_net::mutate_weight(int id)
{
	unsigned int i;

	for(i=0; i<connection_vector.size(); i++)
	{
		if(connection_vector[i].id == id)
		{
			connection_vector[i].weight -= 0.5f;
			connection_vector[i].weight += getRandomFloat(0.0f, 1.0f);			
			break;
		}
	}
}

//---------------------------------------------------------------------
// print_attributes() - prints the main attributes of the net
//---------------------------------------------------------------------
void neural_net::print_attributes()
{
	cout << "Net ID:" << id;
	cout << " Tag:" << tag;
	cout << " Gen:" << generation;
	cout << " Nodes:" << node_vector.size();
	cout << " Connections:" << connection_vector.size();
	cout << endl;
}

//---------------------------------------------------------------------
// print_nodes() - prints the attributes of each node in the net
//---------------------------------------------------------------------
void neural_net::print_nodes()
{
   vector<node>::iterator p;

   for( p = node_vector.begin(); p != node_vector.end(); p++ )
   {
		cout << "Node ID:" << p->id;
		cout << " Tag:" << p->tag;
		cout << " Type:" << p->type;
		cout << " Data:" << p->data;
		cout << endl;
   }
}

//---------------------------------------------------------------------
// print_connections() - prints the attributes of each connection in the net
//---------------------------------------------------------------------
void neural_net::print_connections()
{
   vector<connection>::iterator p;

   for( p=connection_vector.begin(); p!=connection_vector.end(); p++ )
   {
		cout << "Connection ID:" << p->id;
		cout << " Tag:" << p->tag;
		cout << " From:" << p->from;
		cout << " To:" << p->to;
		cout << " Weight:" << p->weight;
		cout << endl;
   }
}

//---------------------------------------------------------------------
// print_input_nodes() - prints attributes of all input nodes 
//---------------------------------------------------------------------
void neural_net::print_input_nodes()
{
	cout << endl << "Input nodes for net " << id << ":" << endl;
	vector<int>::iterator p;
	for( p=input_vector.begin(); p!=input_vector.end(); p++ )
	{
		cout << "node:" << *p << " data:" << node_vector[*p].data;
		cout << endl;
	}
}


//---------------------------------------------------------------------
// print_output_nodes() - prints attributes of all output nodes 
//---------------------------------------------------------------------
void neural_net::print_output_nodes()
{
	cout << "Output nodes for net " << id << ":" << endl;
	vector<int>::iterator p;
	for( p=output_vector.begin(); p!=output_vector.end(); p++ )
	{
		cout << "node:" << *p << " data:" << node_vector[*p].data;
		cout << endl;
	}
}

//---------------------------------------------------------------------
// activate() - single step in neural net
//			 
//---------------------------------------------------------------------
void neural_net::activate()
{
	int i, from, to;
	
	// copy data to temps
	for(i=0; i<input_vector.size(); i++)
	{
		node_vector[i].temp = node_vector[i].data;
	}

	// now activate all connections in order
	for(i=0; i<connection_vector.size(); i++)
	{
		to=connection_vector[i].to;
		from=connection_vector[i].from;

		node_vector[to].temp += node_vector[from].temp *
								connection_vector[i].weight;
	}

	// apply activation function
	for(i=0; i<node_vector.size(); i++)
	{
		if(node_vector[i].type!='i')
		{
			switch(ACTIVATION_FUNCTION)
			{
				// bipolar sigmoid
				case 0:  node_vector[i].temp = (1 - exp(-node_vector[i].temp)) /
											   (1 + exp(-node_vector[i].temp));
						break;

				// sin
				case 1: node_vector[i].temp = sin(node_vector[i].temp);
						break;
				
				// cos
				case 2: node_vector[i].temp = cos(node_vector[i].temp);
						break;

				// tan
				case 3: node_vector[i].temp = tan(node_vector[i].temp);
						break;

				// hyperbolic
				case 4: node_vector[i].temp = (exp(node_vector[i].temp) - exp(-node_vector[i].temp)) /
											  (exp(node_vector[i].temp) + exp(-node_vector[i].temp));
						break;

				// ramp
				case 5: if(node_vector[i].temp>1) node_vector[i].temp=1.0f;
						else if(node_vector[i].temp<-1) node_vector[i].temp=-1.0f;
						break;
				
				// step
				case 6: if(node_vector[i].temp>0) node_vector[i].temp=1.0f;
						else node_vector[i].temp=-1.0f;
						break;

				// gaussian
				case 7: float one_over_sqrt_2pi = 1.0f / sqrt(2.0f*3.14159265);
						node_vector[i].temp = one_over_sqrt_2pi * exp(-1.0f * node_vector[i].temp * node_vector[i].temp / 2.0f);
						break;
			}
		}
	}

	// save all temps to data
	for(i=0; i<node_vector.size(); i++)
	{
		node_vector[i].data=node_vector[i].temp;
	}
}

//---------------------------------------------------------------------
// complexify() - add node and/or link
//---------------------------------------------------------------------
void neural_net::complexify()
{
	int random = getRandomInt(0, 100);
	int i;
	int input, inner, output;
	bool exists = false;

	// pick a random input, inner, and outer
	input = getRandomInt(0, input_vector.size()-1); 
	inner = getRandomInt(input_vector.size()+output_vector.size(), node_vector.size()-1);
	output = getRandomInt(input_vector.size(), input_vector.size()+output_vector.size()-1);

	if(random<40) // add 2 conections
	{
		// check if input>>inner connection exists
		for(i=0; i<connection_vector.size(); i++)
		{
			if(connection_vector[i].from==input &&
			   connection_vector[i].to==inner)
			{
				exists = true;
			}
		}

		// add if possible
		if(!exists) 
		{
			connection new_connection(connection_vector.size(),
									  "new_con",
									  input,
									  inner,
									  -0.5+getRandomFloat(0,1));
			connection_vector.push_back( new_connection );
			exists = false;
		}

		// check if input>>inner connection exists
		for(i=0; i<connection_vector.size(); i++)
		{
			if(connection_vector[i].from==inner &&
			   connection_vector[i].to==output)
			{
				exists = true;
			}
		}

		// add if possible
		if(!exists) 
		{
			connection new_connection(connection_vector.size(),
									  "new_con",
									  inner,
									  output,
									  -0.5+getRandomFloat(0,1));
			connection_vector.push_back( new_connection );
			exists=false;
		}
	}
	else if (random<60) // add new node between input and outer
	{
		// add node
		node N(node_vector.size(), "new_node", 'n', 0.0f);
		node_vector.push_back( N );

		// add connection 1
		connection C1(connection_vector.size(),
					  "new_con",
					   input,
					   N.id,
					   -0.5+getRandomFloat(0,1));
		connection_vector.push_back( C1 );

		// add connection 2
		connection C2(connection_vector.size(),
					  "new_con",
					  N.id,
					  output,
					  -0.5+getRandomFloat(0,1));
		connection_vector.push_back( C2 );
	}
	else if(random < 80) // add new node between input and inner
	{
		// add node
		node N(node_vector.size(), "ni", 'n', 0.0f);
		node_vector.push_back( N );

		// add connection 1
		connection C1(connection_vector.size(),
					  "new_con",
					   input,
					   N.id,
					   -0.5+getRandomFloat(0,1));
		connection_vector.push_back( C1 );

		// add connection 2
		connection C2(connection_vector.size(),
					  "new_con",
					  N.id,
					  inner,
					  -0.5+getRandomFloat(0,1));
		connection_vector.push_back( C2 );
	}
	else // add node between inner and output
	{
		// add node
		node N(node_vector.size(), "no", 'n', 0.0f);
		node_vector.push_back( N );
			// add connection 1
		connection C1(connection_vector.size(),
					  "new_con",
					   inner,
					   N.id,
					   -0.5+getRandomFloat(0,1));
		connection_vector.push_back( C1 );

		// add connection 2
		connection C2(connection_vector.size(),
					  "new_con",
					  N.id,
					  output,
					  -0.5+getRandomFloat(0,1));
		connection_vector.push_back( C2 );
	}
}

//---------------------------------------------------------------------
// draw() - visual representation of a net
//---------------------------------------------------------------------
void neural_net::draw()
{
	vector<D3DXVECTOR3>position_list;

	unsigned int i;
	float top = 4.0f;
	float mid = 0.0f;
	float bottom = -4.0f;
	float left = 4.0f;
	float right = 12.0f;
	float offset = (right-left)/node_vector.size();
	glPointSize(8);
	int num_input=0;
	int num_inner=0;
	int num_output=0;

	// generate a position for each node
	for(i=0; i<node_vector.size(); i++)
	{
		D3DXVECTOR3 V(0.0f, 0.0f, 0.0f);
		
		if(node_vector[i].type=='i')
		{
			V.x=left + num_input*offset;
			V.y=top;
			num_input++;
		}
		else if(node_vector[i].type=='n')
		{
			if(node_vector[i].tag=="ni")
			{
				V.x=left + num_inner*offset;
				V.y= 2.0f;
				num_inner++;
			}
			else if(node_vector[i].tag=="no")
			{
				V.x=left + num_inner*offset;
				V.y= -2.0f;
				num_inner++;
			}
			else
			{
				V.x=left + num_inner*offset;
				V.y= 0.0f;
				num_inner++;
			}
			/*
			V.x=left + num_inner*offset;
			V.y= 0.0f;
			num_inner++;
			*/
		}
		else
		{
			V.x=left+ num_output*offset;
			V.y=bottom;
			num_output++;
		}
		position_list.push_back(V);
	}

	// draw lines
	glColor3f(1.0f, 1.0f, 0.0f);
	for(i=0; i<connection_vector.size(); i++)
	{
		glBegin(GL_LINES);
		glVertex3fv(position_list[connection_vector[i].from]);
		glVertex3fv(position_list[connection_vector[i].to]);
		glEnd();
	}

	for(i=0; i<position_list.size(); i++)
	{
		if(node_vector[i].type=='i') glColor3f(1.0f, 0.0, 0.0f);
		else if(node_vector[i].type=='n') glColor3f(0.0f, 1.0, 0.0f);
		else glColor3f(0.0f, 0.0, 1.0f);

		glBegin(GL_POINTS);
			glVertex3fv(position_list[i]);
		glEnd();
	}
}

#endif
