//-------------------------------------------------------------
// particle_system.h:  particle_system class - a group of particles
//-------------------------------------------------------------

#ifndef PARTICLE_SYSTEM_H
#define PARTICLE_SYSTEM_H

//-------------------------------------------------------------
// includes / defines
//-------------------------------------------------------------
#include "Particle.h"			// particle class


//-------------------------------------------------------------
// external variables
//-------------------------------------------------------------
extern D3DXVECTOR3 RIGHT; 				// right
extern D3DXVECTOR3 UP;					// up
extern float ACCELERATION;				// speed of particles
extern int GENERATION_SHAPE;			// generation shape
extern vector<CPPN> CPPN_VECTOR;	// neural nets for particle systems
extern int PHYSICS_TYPE;				// physics model
extern int PARTICLE_DENSITY;			// particle density

//-------------------------------------------------------------
// particle_system - a collection of particles in 3d space
//-------------------------------------------------------------
class particle_system
{
	public:

	D3DXVECTOR3 position;					// position
	float size;								// size of particle effect
	vector<particle> particle_vector;		// list of particles

	particle_system();						// default contructor
	~particle_system();						// destructor

	void update(float time_elapsed, int index);			
	void draw();
	void reset(int index);
};

//-------------------------------------------------------------
// particle_system::particle_system() - constructor
//-------------------------------------------------------------
particle_system::particle_system()
{
	int i = 0;

	position.x = 0.0f;	
	position.y = 0.0f;
	position.z = 0.0f;
		
	// create some random particles
	for(i=0; i<PARTICLE_DENSITY; i++)				
	{

		particle p(-1.0f + getRandomFloat( 0.0f, 2.0f), // pos x
				   -1.0f + getRandomFloat( 0.0f, 2.0f), // pos y
				   -1.0f + getRandomFloat( 0.0f, 2.0f), // pos z
				   -2.0f + getRandomFloat( 0.0f, 4.0f), // vel x
				   -2.0f + getRandomFloat( 0.0f, 4.0f), // vel y
				   -2.0f + getRandomFloat( 0.0f, 4.0f), // vel z
				   getRandomFloat( 0.0f,  1.0f),		// r
				   getRandomFloat( 0.0f,  1.0f),		// g
				   getRandomFloat( 0.0f,  1.0f),		// b
				   getRandomFloat( 1.0f,  5.0f),		// ttl
				   getRandomFloat( 0.5f, 1.0f));		// size 

		particle_vector.push_back(p);
	}
}

//-------------------------------------------------------------
// particle_system::~particle_system() - destructor
//-------------------------------------------------------------
particle_system::~particle_system()
{}

//-------------------------------------------------------------
// particle_system::update()
//-------------------------------------------------------------
void particle_system::update(float time_elapsed, int index)
{
	unsigned int i;
	D3DXVECTOR3 dtt;

	// inputs = position/DFC 
	// in: 0=px, 1=py, 2=pz, 3=dfc
	// out: 4=vx, 5=vy, 6=vz, 7=r, 8=g, 9=b
	for(i=0; i<particle_vector.size(); i++)
	{
		// update time
		particle_vector[i].time += time_elapsed;

		// plug particle postion info into neural net
		CPPN_VECTOR[index].nodes[0].data = particle_vector[i].position.x; 
		CPPN_VECTOR[index].nodes[1].data = particle_vector[i].position.y; 
		CPPN_VECTOR[index].nodes[2].data = particle_vector[i].position.z; 

		// plug in distance from center
		CPPN_VECTOR[index].nodes[3].data = D3DXVec3Length(&particle_vector[i].position);

		// activate neural net
		CPPN_VECTOR[index].activate();

		// get color info from neural net output
		particle_vector[i].color.x = abs(CPPN_VECTOR[index].nodes[4].data);
		particle_vector[i].color.y = abs(CPPN_VECTOR[index].nodes[5].data);
		particle_vector[i].color.z = abs(CPPN_VECTOR[index].nodes[6].data);

		// get velocity info from neural net output
		particle_vector[i].velocity.x = CPPN_VECTOR[index].nodes[7].data;
		particle_vector[i].velocity.y = CPPN_VECTOR[index].nodes[8].data;
		particle_vector[i].velocity.z = CPPN_VECTOR[index].nodes[9].data;
	
	}// end for

	// move particles based on velocity
	for(i=0; i<particle_vector.size(); i++)
	{
		particle_vector[i].position += particle_vector[i].velocity * time_elapsed * ACCELERATION;
	}

	// reset particle if it has expired
	for(i=0; i<particle_vector.size(); i++)
	{
		// update position, velocity, and color
		if(particle_vector[i].time>particle_vector[i].time_to_live)
		{
			// these attributes do not depend on generation shape
			particle_vector[i].time=0.0f;
			particle_vector[i].velocity.x= 0.0f;  // vel x
			particle_vector[i].velocity.y= 0.0f;  // vel y
			particle_vector[i].velocity.z= 0.0f;  // vel z
			particle_vector[i].color.x=0.0f; // r
			particle_vector[i].color.y=0.0f; // g
			particle_vector[i].color.z=0.0f; // b
			particle_vector[i].time_to_live=getRandomFloat( 1.0f,  5.0f); // ttl
			particle_vector[i].size=getRandomFloat( 0.5f,  1.0f);	// size

			// apply proper generation shape
			switch(GENERATION_SHAPE)
			{
				// tight cloud
				case 0:	particle_vector[i].position.x= -2.0f + getRandomFloat( 0.0f, 4.0f);  // pos x
						particle_vector[i].position.y= -2.0f + getRandomFloat( 0.0f, 4.0f);  // pos y
						particle_vector[i].position.z= -2.0f + getRandomFloat( 0.0f, 4.0f);  // pos z 
						break;

				// point
				case 1:	particle_vector[i].position.x= -0.2f + getRandomFloat( 0.0f, 0.4f);  // pos x
						particle_vector[i].position.y= -0.2f + getRandomFloat( 0.0f, 0.4f);  // pos y
						particle_vector[i].position.z= -0.2f + getRandomFloat( 0.0f, 0.4f);  // pos z 
						break;

				// line
				case 2:	particle_vector[i].position.x= position.x;  // pos x
						particle_vector[i].position.y= position.y;  // pos y
						particle_vector[i].position.z= -3.0f + getRandomFloat( 0.0f, 6.0f);  // pos z 
						break;

				// circle
				case 3:	particle_vector[i].position.x= -3.0f + getRandomFloat( 0.0f, 6.0f);  // pos x
						particle_vector[i].position.y= position.y;							 // pos y
						particle_vector[i].position.z= -3.0f + getRandomFloat( 0.0f, 6.0f);  // pos z 
						D3DXVec3Normalize(&particle_vector[i].position, &particle_vector[i].position);
						particle_vector[i].position*=3.0f;
						break;
				
				// wide cloud
				case 4:	particle_vector[i].position.x= -4.0f + getRandomFloat( 0.0f, 8.0f);  // pos x
						particle_vector[i].position.y= -4.0f + getRandomFloat( 0.0f, 8.0f);  // pos y
						particle_vector[i].position.z= -4.0f + getRandomFloat( 0.0f, 8.0f);  // pos z 
						break;
				
				// field
				case 5:	particle_vector[i].position.x= -10.0f + getRandomFloat( 0.0f, 20.0f);  // pos x
						particle_vector[i].position.y= -10.0f + getRandomFloat( 0.0f, 20.0f);  // pos y
						particle_vector[i].position.z= -10.0f + getRandomFloat( 0.0f, 20.0f);  // pos z 
			} // end switch
		}// end if
	}// end for
}

//-------------------------------------------------------------
// particle_system::draw()
//-------------------------------------------------------------
void particle_system::draw()
{
	unsigned int i;
	D3DXVECTOR3 A,B,C,D;	// corners of billboard

	// iterate through particle list ...
	for(i=0; i<particle_vector.size(); i++)		
	{		
		// compute corners as vectors A,B,C,D
		A = particle_vector[i].position -				
			(RIGHT+UP) *
			particle_vector[i].size;	
		B = particle_vector[i].position +
			(RIGHT-UP) *
			particle_vector[i].size;
		C = particle_vector[i].position +
			(RIGHT+UP) * 
			particle_vector[i].size;
		D = particle_vector[i].position -
			(RIGHT-UP) *
			particle_vector[i].size;

		glColor3f(particle_vector[i].color.x,			
				  particle_vector[i].color.y,
				  particle_vector[i].color.z);		

		glBegin(GL_TRIANGLE_STRIP);		// render billboard

			glTexCoord2d(1,1); 
			glVertex3fv(C);
			glTexCoord2d(0,1); 
			glVertex3fv(D);
			glTexCoord2d(1,0); 
			glVertex3fv(B);
			glTexCoord2d(0,0); 
			glVertex3fv(A);
		glEnd();
	}	
}

//-------------------------------------------------------------
// particle_system::reset() - resets to simple system
//-------------------------------------------------------------
void particle_system::reset(int index)
{
	// reset CPPN
	CPPN_VECTOR[index].reset(4,6);
}

#endif
