//-------------------------------------------------------------
// trail_system.h: trail_system class - a group of particles
//-------------------------------------------------------------

#ifndef TRAIL_SYSTEM_H
#define TRAIL_SYSTEM_H

//-------------------------------------------------------------
// includes / defines
//-------------------------------------------------------------

#include <d3dx9.h>				// DirectX 9
#include <vector>				// vector class
#include "TrailParticle.h"		// particle class
#include "CPPN.h"				// neural net class
#include "Random.h"				// random number functions
using namespace std;

//-------------------------------------------------------------
// trail_system - a single particle in 3d space
//-------------------------------------------------------------
class trail_system
{
	public:

	D3DXVECTOR3 position;					// position
	float size;
	vector<trail_particle> particle_vector;	// trail particles
	float frequency;						// how often particles drop

	trail_system();							// default contructor
	~trail_system();						// destructor

	void update(float time_elapsed, int index);			
	void draw();
	void reset(int index);
};

//-------------------------------------------------------------
// trail_system::trail_system() - constructor
//-------------------------------------------------------------
trail_system::trail_system()
{
	position.x=0.0f;
	position.y=0.0f;
	position.z=0.0f;
	frequency=0.25f;

	int i;
	for(i=0; i<PARTICLE_DENSITY; i++)
	{
		trail_particle t;
		particle_vector.push_back(t);
	}
}

//-------------------------------------------------------------
// trail_system::~trail_system() - destructor
//-------------------------------------------------------------
trail_system::~trail_system()
{}

//-------------------------------------------------------------
// trail_system::update()
//-------------------------------------------------------------
void trail_system::update(float time_elapsed, int index)
{
	// local variables
	int i,j, front;
	float decrement;

	// update all particles
	for(i=0; i<particle_vector.size(); i++)
	{
		// update timers
		particle_vector[i].time += time_elapsed;
		particle_vector[i].drop_timer += time_elapsed;

		// check if particle expired
		if(particle_vector[i].time > particle_vector[i].time_to_live)
		{
			// apply proper generation shape on reset
			switch(GENERATION_SHAPE)
			{
				// tight cloud
				case 0:	particle_vector[i].position.x= -2.0f + getRandomFloat( 0.0f, 4.0f);  // pos x
						particle_vector[i].position.y= -2.0f + getRandomFloat( 0.0f, 4.0f);  // pos y
						particle_vector[i].position.z= -2.0f + getRandomFloat( 0.0f, 4.0f);  // pos z 
						break;
				// point
				case 1:	particle_vector[i].position.x= -0.5f + getRandomFloat( 0.0f, 1.0f);  // pos x
						particle_vector[i].position.y= -0.5f + getRandomFloat( 0.0f, 1.0f);  // pos y
						particle_vector[i].position.z= -0.5f + getRandomFloat( 0.0f, 1.0f);  // pos z 
						break;
				// line
				case 2:	particle_vector[i].position.x= position.x;  // pos x
						particle_vector[i].position.y= position.y;  // pos y
						particle_vector[i].position.z= -3.0f + getRandomFloat( 0.0f, 6.0f);  // pos z 
						break;
				// circle
				case 3:	particle_vector[i].position.x= -3.0f + getRandomFloat( 0.0f, 6.0f);  // pos x
						particle_vector[i].position.y= position.y;							 // pos y
						particle_vector[i].position.z= -3.0f + getRandomFloat( 0.0f, 6.0f);  // pos z 
						D3DXVec3Normalize(&particle_vector[i].position, &particle_vector[i].position);
						particle_vector[i].position*=3.0f;
						break;
				// wide cloud
				case 4:	particle_vector[i].position.x= -4.0f + getRandomFloat( 0.0f, 8.0f);  // pos x
						particle_vector[i].position.y= -4.0f + getRandomFloat( 0.0f, 8.0f);  // pos y
						particle_vector[i].position.z= -4.0f + getRandomFloat( 0.0f, 8.0f);  // pos z 
						break;
				// field
				case 5:	particle_vector[i].position.x= -10.0f + getRandomFloat( 0.0f, 20.0f);  // pos x
						particle_vector[i].position.y= -10.0f + getRandomFloat( 0.0f, 20.0f);  // pos y
						particle_vector[i].position.z= -10.0f + getRandomFloat( 0.0f, 20.0f);  // pos z 
						break;
			} // end switch
			particle_vector[i].color.x=0.0f;
			particle_vector[i].color.y=0.0f;
			particle_vector[i].color.z=0.0f;
			particle_vector[i].time=0.0f;
			particle_vector[i].drop_timer=0.0f;
			particle_vector[i].time_to_live = getRandomFloat(5.0f, 10.0f);
			particle_vector[i].front=0;
		}

		// check to drop new particles
		if(particle_vector[i].drop_timer > frequency)
		{
			// where particle is dropped
			front = particle_vector[i].front;

			// drop the particle
			particle_vector[i].position_vector[front] = particle_vector[i].position;
			particle_vector[i].size_vector[front] = particle_vector[i].size;

			// move front forward
			particle_vector[i].front++;
			if(particle_vector[i].front==particle_vector[i].position_vector.size())
			{
				particle_vector[i].front=0;	
			}

			// reset drop timer
			particle_vector[i].drop_timer = 0.0f;
		}

		// decrement all trail sizes
		decrement = 1.5f * time_elapsed / particle_vector[i].time_to_live;
		for(j=0; j<particle_vector[i].size_vector.size(); j++)
		{
			particle_vector[i].size_vector[j] -= decrement;
		}

		// plug particle postion info into neural net
		CPPN_VECTOR[index].nodes[0].data = particle_vector[i].position.x; 
		CPPN_VECTOR[index].nodes[1].data = particle_vector[i].position.y; 
		CPPN_VECTOR[index].nodes[2].data = particle_vector[i].position.z; 

		// plug in distance from center
		CPPN_VECTOR[index].nodes[3].data = D3DXVec3Length(&particle_vector[i].position);

		// activate neural net
		CPPN_VECTOR[index].activate();

		// get color info from neural net output
		particle_vector[i].color.x = abs(CPPN_VECTOR[index].nodes[4].data);
		particle_vector[i].color.y = abs(CPPN_VECTOR[index].nodes[5].data);
		particle_vector[i].color.z = abs(CPPN_VECTOR[index].nodes[6].data);

		// get velocity info from neural net output
		particle_vector[i].velocity.x = CPPN_VECTOR[index].nodes[7].data;
		particle_vector[i].velocity.y = CPPN_VECTOR[index].nodes[8].data;
		particle_vector[i].velocity.z = CPPN_VECTOR[index].nodes[9].data;
		
		// update the trail particle position
		particle_vector[i].position += particle_vector[i].velocity * 
									   time_elapsed * 
									   ACCELERATION;
	}
}

//-------------------------------------------------------------
// particle_system::draw()
//-------------------------------------------------------------
void trail_system::draw()
{
	// temp variables
	unsigned int i,j;
	D3DXVECTOR3 A,B,C,D, pos;
	float s;

	// iterate through particle list ...
	for(i=0; i<particle_vector.size(); i++)		
	{	
		// get size, decrement
		s = particle_vector[i].size;

		// draw the leading particle
		pos = particle_vector[i].position;
		glColor3fv(particle_vector[i].color);	
		
		// compute corners as vectors A,B,C,D
		A = pos - (RIGHT+UP) * s;	
		B = pos + (RIGHT-UP) * s;
		C = pos + (RIGHT+UP) * s;
		D = pos - (RIGHT-UP) * s;

		glBegin(GL_TRIANGLE_STRIP);		// render billboard
			glTexCoord2d(1,1); 
			glVertex3f(C.x,C.y,C.z);
			glTexCoord2d(0,1); 
			glVertex3f(D.x,D.y,D.z);
			glTexCoord2d(1,0); 
			glVertex3f(B.x,B.y,B.z);
			glTexCoord2d(0,0); 
			glVertex3f(A.x,A.y,A.z);
		glEnd();

		// draw each trail particle
		for(j=0; j<particle_vector[i].position_vector.size(); j++)
		{
			// only draw particles that are visible
			if(particle_vector[i].size_vector[j] > 0.0f)
			{
				// get particle position and color
				pos = particle_vector[i].position_vector[j];
				s = particle_vector[i].size_vector[j];			

				// compute corners as vectors A,B,C,D
				A = pos - (RIGHT+UP) * s;	
				B = pos + (RIGHT-UP) * s;
				C = pos + (RIGHT+UP) * s;
				D = pos - (RIGHT-UP) * s;

				glBegin(GL_TRIANGLE_STRIP);		// render billboard
					glTexCoord2d(1,1); 
					glVertex3f(C.x,C.y,C.z);
					glTexCoord2d(0,1); 
					glVertex3f(D.x,D.y,D.z);
					glTexCoord2d(1,0); 
					glVertex3f(B.x,B.y,B.z);
					glTexCoord2d(0,0); 
					glVertex3f(A.x,A.y,A.z);
				glEnd();
			}
		}
	}
}

//-------------------------------------------------------------
// particle_system::reset() - resets to simple system
//-------------------------------------------------------------
void trail_system::reset(int index)
{
	// reset CPPN
	CPPN_VECTOR[index].reset(4,6);
}

#endif
